-- array to hold debug messages
text = {}
textr = {}

HC = require 'hardoncollider'

-- Object class definition
Object = {id=0,image=nil,left=0,top=0,bottom=0,right=0,dragging={active=false,diffX=0,diffY=0},deletion=false,collider=nil}

function Object:new (o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self
  return o
end

function Object:setRnB()
	self.right=self.left+self.image:getWidth()
	self.bottom=self.top+self.image:getHeight()
end

function Object:moveTo(x,y)
	self.left = x
	self.top = y

	-- move the collider
	local cx,cy = getObjectCenter(self)
	self.collider:moveTo(cx,cy)
	
	self.right=self.left+self.image:getWidth()
	self.bottom=self.top+self.image:getHeight()
end

-- end class

function love.load()
	--love.graphics.setMode(0,0)
    --love.graphics.toggleFullscreen()
	
	-- initialize collider library
    Collider = HC(100, on_collision, collision_stop)

	dragStartX = -1
	dragStartY = -1
	
	love.graphics.setBackgroundColor(100,100,100)
	
	ubound_objects = 0 -- upper boundary of the displayed objects table
	objects = {} -- displayed objects table
	selectedobjects = {} -- objects that have been selected
	draggedobjectid = 0
	objectsfordeletion = false
	
	initialiseobjecttable() -- create the table of available objects
	
	createmenu()
	
	testshape = nil -- global for testing
end

function love.draw()
    -- DEBUG: print messages
    printDebug()
	printDebugR()

	love.graphics.print("F for full screen.", 100, love.graphics.getHeight() - 50)
	--love.graphics.print(love.mouse.getX() .. "," .. love.mouse.getY(),10,love.graphics.getHeight()-150)
	
	drawmenu()

	-- draw the bin
	newimage = love.graphics.newImage("images/trash.png")
	love.graphics.draw(newimage,20,(love.graphics.getHeight()-(newimage:getHeight()+20)))

	-- loop through the objects and draw them
	for object in iter(objects) do
		-- draw an object
		love.graphics.draw(object.image, object.left, object.top)
		
		--local positions = "left=" .. object.left .. "\ntop=" .. object.top .. "\nright=" .. object.right .. "\nbottom=" .. object.bottom .. "\nwidth=" .. object.image:getWidth() .. "\nheight=" .. object.image:getHeight()
		--love.graphics.print(positions, object.right+1, object.top)
    end
	
	-- highlight selected objects
	for selectedobject in iter(selectedobjects) do
		selectedobject.object.collider:draw('line') -- shows the collision components
		--love.graphics.draw(object.image, object.left, object.top)
	end
end

function love.update()
	-- loop through menus then objects and if one is set to dragging = true, update it's left and top to the mouse cursor position
	if draggedobjectid ~= 0 then
		object = objects[draggedobjectid]
		if object.dragging.active == true then
			object:moveTo(love.mouse.getX() - object.dragging.diffX, love.mouse.getY() - object.dragging.diffY)
		end
	end
	
	-- deletion check
	if objectsfordeletion==true then
		-- remove all objects marked for deletion by copying the table to a temp table then copying back only those not marked for deletion
		tempobjects = objects
		objects = {}
		ubound_objects = 0
		for object in iter(tempobjects) do
			if object.deletion==false then -- copy to the objects table if deletion is false
				addobject(object)
			end
		end
		objectsfordeletion = false
		selectedobjects = {}
	end
	
    -- update collisions
    Collider:update(dt)
end

function love.mousepressed(x,y,button)
	handleobjectclicks(x,y,button)
	handlemenuclicks(x,y,button)
	
	-- deselect all selected and then select the clicked object
	if button == "l" then
		selectObject(objects[draggedobjectid])
	elseif button =="r" then
		-- right button click
	end
end

function love.mousereleased(x,y,button)
	-- if released on the menu, delete the object
	-- otherwise cancel the current drag
	if draggedobjectid ~= 0 then
		object = objects[draggedobjectid]
		if object.dragging.active == true then
			-- check if it's being dropped on the menu
			if object.right < (love.graphics.getWidth()-104) then
				object.dragging.active = false
			else
				-- move object out of the menu
				object.left = love.graphics.getWidth() - (104 + object.image:getWidth()+1)
			end
			
			-- check if it's dropped on the trash
			if not(x < 20 or x > 84 or y < (love.graphics.getHeight()-84) or y > (love.graphics.getHeight()-20)) then
				object.deletion=true
				objectsfordeletion=true
			end
			
			-- snap the object if necessary.
			snapObject(object)
			
			-- reset the dragging id
			draggedobjectid = 0
			dragStartX = -1
			dragStartY = -1
		end
	end
end

function love.keyreleased(key)
	if key == "delete" then
		-- delete selected objects
		for selectedobj in iter(selectedobjects) do
			selectedobj.object.deletion=true
			objectsfordeletion = true
		end
		
		--selectedobjects = {}
	elseif key == "f" then
		love.graphics.setMode(0,0)
		love.graphics.toggleFullscreen()
		createmenu()
	elseif key == "escape" then
		love.event.push("quit")   -- actually causes the app to quit
	end
end

-- this is called when two shapes collide
function on_collision(dt, shape_a, shape_b, mtv_x, mtv_y)
    --if draggedobjectid ~= 0 then
	--	object = objects[draggedobjectid]
	--	object.colliding = true
	--	object.collidingwith = shape_b
		
	--	addDebug("Colliding")
	--end
end

-- this is called when two shapes stop colliding
function collision_stop(dt, shape_a, shape_b)
	--if draggedobjectid ~= 0 then
	--	object = objects[draggedobjectid]
	--	object.colliding = false
	--	object.collidingwith = nil
	--addDebug("stopped Colliding")
	--end
end

-- table of available components
function initialiseobjecttable()
	-- table of available objects
	objectsavailable = {}
	objectsavailable[1] = "images/empty_64_64.png"
	objectsavailable[2] = "images/engine_64_64.png"
	objectsavailable[3] = "images/laser_64_64.png"
	objectsavailable[4] = "images/bridge_64_64.png"
	objectsavailable[5] = "images/fuel_cell_mechanical_64_64.png"
--	currentobject = 1 -- current object chosen from those available
--	numobjectsavailable = 2 -- how many objects can be chosen from
end

function createmenu()
	menuobjects = {} -- initialise the table holding the menu objects
	
	-- draw all the available objects one under the other
	-- loop through available objects, create a Object instance and populate then add to the menu objects table
	-- each  object has it's x and y positions offset from the previous object in the table
	for i=1,table.getn(objectsavailable) do
		thisobject = Object:new()
		thisobject.image = love.graphics.newImage(objectsavailable[i])
		if i>1 then
			thisobject.left = menuobjects[i-1].left
			thisobject.top = menuobjects[i-1].top + (thisobject.image:getHeight() + 10)
		else
			thisobject.left = love.graphics.getWidth() - (thisobject.image:getWidth() + 20)
			thisobject.top = 40
		end
		
		thisobject:setRnB() -- sets right and bottom
		
		thisobject.id=i
		
		menuobjects[i] = thisobject
	end

end

function drawmenu()
	-- draw table dependant on screen res settings
	love.graphics.line(love.graphics.getWidth()-104,0,love.graphics.getWidth()-104,love.graphics.getHeight())
	love.graphics.print("MENU", love.graphics.getWidth()-84, 10)
	
	-- draw the menu objects
	for object in iter(menuobjects) do
		love.graphics.draw(object.image, object.left, object.top)
	end	
end

function handlemenuclicks(mouse_x,mouse_y,button)
	if button=="l" then
		-- loop through all objects and see if one was clicked on
		for object in iter(menuobjects) do
			-- check if x,y are in a menu image
			if not(mouse_x < object.left or mouse_x > object.right or mouse_y < object.top or mouse_y > object.bottom) then
				-- if in a menu image, create a new object and copy settings from the menu object
				local newobject = cloneobject(object)

				newobject.left = love.mouse.getX() - 15
				newobject.top = love.mouse.getY() - 15
				newobject:setRnB() -- sets right and bottom
				newobject.dragging.active = true
				newobject.dragging.diffX = mouse_x - newobject.left
				newobject.dragging.diffY = mouse_y - newobject.top
				
				draggedobjectid = addobject(newobject) -- add new object to the objects table
				break
			end
		end	
	end
end

function handleobjectclicks(mouse_x,mouse_y,button)
	if button=="l" then
		-- loop through all objects and see if one was clicked on
		for object in iter(objects) do
			if not(mouse_x < object.left or mouse_x > object.right or mouse_y < object.top or mouse_y > object.bottom) then
				object.dragging.active = true
				object.dragging.diffX = mouse_x - object.left
				object.dragging.diffY = mouse_y - object.top

				dragStartX = object.left
				dragStartY = object.top
				draggedobjectid = object.id
				break
			end
		end	
	end
end

function cloneobject(cloneme)
	-- copy the object passed
	thisobject = Object:new()
	thisobject.image = cloneme.image
	--thisobject.image = love.graphics.newImage("images/heart.jpg") -- for Kristie

	return thisobject
end

function iter(t)
	local i = 0
	local n = table.getn(t)
	return function ()
			i = i + 1
			if i <= n then return t[i] end
		end
end

function addobject(thisobject)
	-- Add the passed object to the objects table
	-- increment the upper boundary then make this the object ID
	ubound_objects = ubound_objects + 1
	thisobject.id = ubound_objects
	
	-- Add collision object
	thisobject.collider = Collider:addRectangle(thisobject.left,thisobject.top,thisobject.image:getWidth(),thisobject.image:getHeight())
	
	-- Add thisobject to the objects table
	objects[ubound_objects] = thisobject
	
	return ubound_objects
end

function getObjectCenter(obj)
	cx = obj.left + (obj.image:getWidth() / 2)
	cy = obj.top + (obj.image:getHeight() / 2)
	
	return cx,cy
end

function addDebug(x)
	text[#text+1] = os.time() .. ': ' ..x
end

function printDebug()
	for i=1, #text do
		love.graphics.print(text[i] .. '\n', 10, (10*i))
	end
end

function addDebugR(x)
	textr[#textr+1] = os.time() .. ': ' .. x
end

function printDebugR()
	for i=0, (#textr-1) do
		love.graphics.print(textr[#textr-i] .. '\n', 10, (10*i))
	end
end

function snapObject(object)
	local colliding = false
	local test_left, test_top, test_right, test_bottom
	local v_mtv_x, v_mtv_y

	--collidingtable = object, mtv_x,  mtv_y
	collidingtable = {}
	
	for otherobject in iter(objects) do
		if otherobject ~= object then -- don't compare to itself
			if object.collider:collidesWith(otherobject.collider) then -- check if these two objects collide
				local bool,x,y = object.collider:collidesWith(otherobject.collider) -- check if these two components collide
				colliding = true
				-- add the object to the colliding table. store the object and the move to vectors
				table.insert(collidingtable, {object = otherobject, mtv_x = x, mtv_y = y})
			end
		end
	end

	if colliding==true then
		if #collidingtable < 3 then
			objectForSnap = 1 -- the object we'll move to at the end
			i = 1 -- count # objects
			for collidingobj in iter(collidingtable) do
				-- each colliding table entry consists of an object and move-to-vectors
				if mouseOver(collidingobj.object) then -- if mouse is over this object, use that as the moving base
					objectForSnap = i
				end
				i = i + 1
			end
			-- if mouse is over no object, use the first collided with
			
			-- move to the positions stored
			v_mtv_x = collidingtable[objectForSnap].mtv_x
			v_mtv_y = collidingtable[objectForSnap].mtv_y
			object:moveTo(object.left + v_mtv_x, object.top + v_mtv_y)
		else
			-- handle collisions with 3 or more objects differently
			movemade=false
			for objtemp1 in iter(collidingtable) do
				for objtemp2 in iter(collidingtable) do
					if not(objtemp1.object.top==objtemp2.object.top or objtemp1.object.left==objtemp2.object.left or objtemp1.object.right==objtemp2.object.right or objtemp1.object.bottom==objtemp2.object.bottom) then
						if objtemp1.mtv_x ~= 0 then
							object:moveTo(object.left + objtemp1.mtv_x, object.top + objtemp2.mtv_y)
						else
							object:moveTo(object.left + objtemp2.mtv_x, object.top + objtemp1.mtv_y)
						end
						movemade = true
						break
					end
				end

				if movemade then
					break
				end
			end
		end
		
		-- check if it's colliding with anything else and if TRUE move back to the original position
		if collidingWithObjects(object,collidingtable[objectForSnap].object) then -- object to test and object to ignore
			-- if dragStartX and dragStartY are -1, this came from the menu. delete it
			if dragStartX == -1 and dragStartY == -1 then
				object.deletion=true
				objectsfordeletion=true
			else
				object:moveTo(dragStartX, dragStartY)
			end
		end
	end
	
	object:moveTo(snapToGrid(object.left, object.top, 16))
end

function getNewPositions(shape_a, shape_b)
	-- shape_a is the test shape
	-- shape_b is the shape that shape_a collided with
	
	local current = "top" -- default
	
	a_x,a_y,a_x2,a_y2 = shape_a:bbox()
	b_x,b_y,b_x2,b_y2 = shape_b:bbox()
	
	if a_x >= b_x2 then
		-- shape is to the right of the colliding shape so move to bottom
		return b_x,(b_y2)
	elseif a_x2 <= b_x then
		-- shape is to the left of the colliding shape so move to top
		return b_x,((b_y)-(a_y2-a_y))
	elseif a_y >= b_y2 then
		-- shape is below the colliding shape so move to left
		return ((b_x)-(a_x2-a_x)),(b_y)
	elseif a_y2 <= b_y then
		-- shape is above the colliding shape so move to right
		return (b_x2),b_y
	else
		return 0,0
	end
end

-- return x and y snapped to a grid of (size * size)
function snapToGrid(x,y,size)
	local snap_pos_x, snap_pos_y
	snap_pos_x = math.floor(x / size) * size
	snap_pos_y = math.floor(y / size) * size
	
	return snap_pos_x, snap_pos_y
end

function mouseOver(object)
	return not(love.mouse.getX() < object.left or love.mouse.getX() > object.right or love.mouse.getY() < object.top or love.mouse.getY() > object.bottom)
end

-- check is this object is colliding with other objects. Returns bool,x,y
-- Accepts an object. the optional parameter is an object to ignore
function collidingWithObjects(object, ignore)
	local bool = false
	local x = -1
	local y = -1

	if object and ignore then -- if both args passed
		for otherobject in iter(objects) do
			 -- don't compare to itself or the ignore object
			if otherobject ~= object and otherobject ~= ignore then
				if object.collider:collidesWith(otherobject.collider) then -- check if these two objects collide
					if justEdgesColliding(object, otherobject) == false then
						bool,x,y = object.collider:collidesWith(otherobject.collider) -- check if these two components collide
						break
					end
				end
			end
		end
	else -- if only one arg passed
		bool = false
	end

	--if bool==true then addDebug("True") else addDebug("False") end
	
	return bool,x,y
end

-- check if it is just the edges colliding
function justEdgesColliding(obj1, obj2)
	return ((obj1.right==obj2.left) or (obj1.left==obj2.right) or (obj1.bottom==obj2.top) or (obj1.top==obj2.bottom))
end

function selectObject(obj)
	local alreadyselected = false
	
	if obj ~= nil then
		-- clear all selections if ctrl is not held down
		if not love.keyboard.isDown("rctrl") and not love.keyboard.isDown("lctrl") then
			--addDebug("de-selecting all")
			selectedobjects = {}
		end
		
		-- add this object to the selected objects array if it is not already there
		for selectedobj in iter(selectedobjects) do
			if selectedobj.object == obj then
				--addDebug("already selected")
				alreadyselected = true
			end
		end
		
		if alreadyselected == false then
			--addDebug("selecting this")
			table.insert(selectedobjects, {object = obj})
		end
	else
		--addDebug("de-selecting all")
		selectedobjects = {}
	end
end
